/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors :
 ...
***********************************************************************/
package org.datanucleus.store.odf.fieldmanager;

import java.lang.reflect.Array;
import java.util.Calendar;
import java.util.Collection;
import java.util.Currency;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.fieldmanager.AbstractFieldManager;
import org.datanucleus.store.odf.ODFUtils;
import org.datanucleus.store.types.ObjectLongConverter;
import org.datanucleus.store.types.ObjectStringConverter;
import org.datanucleus.util.NucleusLogger;
import org.openoffice.odf.doc.element.table.OdfTableCell;
import org.openoffice.odf.doc.element.table.OdfTableRow;
import org.openoffice.odf.dom.type.office.OdfValueType;

/**
 * FieldManager to handle the update of information into an ODF spreadsheet row using an object.
 */
public class UpdateFieldManager extends AbstractFieldManager
{
    /** StateManager for object being inserted. */
    private final ObjectProvider sm;

    /** Row being inserted. */
    private final OdfTableRow row;

    public UpdateFieldManager(ObjectProvider sm, OdfTableRow row)
    {
        this.sm = sm;
        this.row = row;
    }

    public void storeBooleanField(int fieldNumber, boolean value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setBooleanValue(value);
    }

    public void storeByteField(int fieldNumber, byte value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setValue(new Double(value));
    }

    public void storeCharField(int fieldNumber, char value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setStringValue("" + value);
    }

    public void storeDoubleField(int fieldNumber, double value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setValue(value);
    }

    public void storeFloatField(int fieldNumber, float value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setValue(new Double(value));
    }

    public void storeIntField(int fieldNumber, int value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setValue(new Double(value));
    }

    public void storeLongField(int fieldNumber, long value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setValue(new Double(value));
    }

    public void storeShortField(int fieldNumber, short value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        cell.setValue(new Double(value));
    }

    public void storeStringField(int fieldNumber, String value)
    {
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
        if (value == null)
        {
            cell.setStringValue(ODFUtils.STRING_NULL_VALUE);
        }
        else
        {
            cell.setStringValue(value);
        }
    }

    public void storeObjectField(int fieldNumber, Object value)
    {
        ExecutionContext ec = sm.getExecutionContext();
        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        AbstractMemberMetaData mmd =
            sm.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        int relationType = mmd.getRelationType(clr);
        int colIndex = ODFUtils.getColumnIndexForFieldOfClass(sm.getClassMetaData(), fieldNumber);
        OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);

        if (value == null)
        {
            if (Number.class.isAssignableFrom(mmd.getType()))
            {
                cell.setValue(null);
            }
            else if (Date.class.isAssignableFrom(mmd.getType()) || Calendar.class.isAssignableFrom(mmd.getType()))
            {
                cell.setDateValue(null);
            }
            else
            {
                // Assume it's a String type TODO Set the type based on the field type
                cell.setValueType(OdfValueType.STRING);
            }
        }
        else if (relationType == Relation.NONE)
        {
            if (value instanceof Calendar)
            {
                try
                {
                    cell.setDateValue(DatatypeFactory.newInstance().newXMLGregorianCalendar((GregorianCalendar)value));
                }
                catch (DatatypeConfigurationException e)
                {
                    throw new NucleusException("Error occurred inserting field with value=" + value, e);
                }
            }
            else if (value instanceof Date)
            {
                try
                {
                    GregorianCalendar cal = (GregorianCalendar)GregorianCalendar.getInstance();
                    cal.setTime((Date)value);
                    cell.setDateValue(DatatypeFactory.newInstance().newXMLGregorianCalendar(cal));
                }
                catch (DatatypeConfigurationException e)
                {
                    throw new NucleusException("Error occurred inserting field with value=" + value, e);
                }
            }
            else if (value instanceof Boolean)
            {
                cell.setBooleanValue((Boolean)value);
            }
            else if (value instanceof Byte)
            {
                cell.setValue(new Double((Byte)value));
            }
            else if (value instanceof Character)
            {
                cell.setStringValue("" + value);
            }
            else if (value instanceof Double)
            {
                cell.setValue((Double)value);
            }
            else if (value instanceof Float)
            {
                cell.setValue((Double)value);
            }
            else if (value instanceof Integer)
            {
                cell.setValue(new Double((Integer)value));
            }
            else if (value instanceof Long)
            {
                cell.setValue(new Double((Long)value));
            }
            else if (value instanceof Short)
            {
                cell.setValue(new Double((Short)value));
            }
            else if (value instanceof Currency)
            {
                ObjectStringConverter converter = 
                    sm.getExecutionContext().getOMFContext().getTypeManager().getStringConverter(value.getClass());
                cell.setCurrency(converter.toString(value));
            }
            else if (value instanceof Enum)
            {
                cell.setStringValue(((Enum)value).name());
            }
            else
            {
                // See if we can persist it as a Long/String
                boolean useLong = false;
                ColumnMetaData[] colmds = mmd.getColumnMetaData();
                if (colmds != null && colmds.length == 1)
                {
                    String jdbc = colmds[0].getJdbcType();
                    if (jdbc != null && (jdbc.equalsIgnoreCase("int") || jdbc.equalsIgnoreCase("integer")))
                    {
                        useLong = true;
                    }
                }
                ObjectStringConverter strConv = 
                    sm.getExecutionContext().getOMFContext().getTypeManager().getStringConverter(value.getClass());
                ObjectLongConverter longConv = 
                    sm.getExecutionContext().getOMFContext().getTypeManager().getLongConverter(value.getClass());
                if (useLong)
                {
                    if (longConv != null)
                    {
                        cell.setValue(Double.valueOf(longConv.toLong(value)));
                        return;
                    }
                }
                else
                {
                    if (strConv != null)
                    {
                        cell.setStringValue(strConv.toString(value));
                        return;
                    }
                    else
                    {
                        cell.setValue(Double.valueOf(longConv.toLong(value)));
                        return;
                    }
                }

                NucleusLogger.PERSISTENCE.warn("Dont currently support persistence of field=" + mmd.getFullFieldName() +
                    " type=" + value.getClass().getName() + " to ODF");
            }
        }
        else if (relationType == Relation.ONE_TO_ONE_BI || relationType == Relation.ONE_TO_ONE_UNI ||
            relationType == Relation.MANY_TO_ONE_BI)
        {
            // Persistable object - persist the related object and store the identity in the cell
            Object valuePC = sm.getExecutionContext().persistObjectInternal(value, sm, fieldNumber, -1);
            Object valueId = sm.getExecutionContext().getApiAdapter().getIdForObject(valuePC);
            cell.setValueType(OdfValueType.STRING);
            cell.setStringValue("[" + valueId.toString() + "]");
        }
        else if (relationType == Relation.MANY_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_BI ||
            relationType == Relation.ONE_TO_MANY_UNI)
        {
            // Collection/Map/Array
            if (mmd.hasCollection())
            {
                // Just overwrite the current cell value
                StringBuffer cellValue = new StringBuffer("[");
                Collection coll = (Collection)value;
                Iterator collIter = coll.iterator();
                while (collIter.hasNext())
                {
                    Object element = collIter.next();
                    Object elementPC = sm.getExecutionContext().persistObjectInternal(element, sm, fieldNumber, -1);
                    Object elementID = sm.getExecutionContext().getApiAdapter().getIdForObject(elementPC);
                    cellValue.append(elementID.toString());
                    if (collIter.hasNext())
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                cell.setValueType(OdfValueType.STRING);
                cell.setStringValue(cellValue.toString());
            }
            else if (mmd.hasMap())
            {
                // TODO Implement maps
            }
            else if (mmd.hasArray())
            {
                // Just overwrite the current cell value
                StringBuffer cellValue = new StringBuffer("[");
                for (int i=0;i<Array.getLength(value);i++)
                {
                    Object element = Array.get(value, i);
                    Object elementPC = sm.getExecutionContext().persistObjectInternal(element, sm, fieldNumber, -1);
                    Object elementID = sm.getExecutionContext().getApiAdapter().getIdForObject(elementPC);
                    cellValue.append(elementID.toString());
                    if (i < (Array.getLength(value)-1))
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                cell.setValueType(OdfValueType.STRING);
                cell.setStringValue(cellValue.toString());
            }
        }
    }
}